set ::THISPROG(NAME) Sticky
set ::THISPROG(VER)  2.0

package provide stickyWM  $::THISPROG(VER)

##  stickywm.tcl
##
##  stickywm.tcl - StickyNote's Window Manager.
##
##  "stickyWM" is the main program.
##  It creates a (sort of) 'control panel' with buttons allowing you to:
##    create a new stickyNote
##    show/hide the stickyNote-template manager.
##
##  stickyWM is also responsible for 
##   * saving all stickies when the app closes,
##   * reloading the saved stickies when the app starts.
##   * hide/show all the stickies when the control panel is iconized/restored.
##   * trap logoff/shutdown (provided tcl-core is enabled)
##   * prevent from starting twice
##   * schedule every 5 minute an auto-save procedure
##     (only if something has changed).
##   * add auto-fading to stickyNotes
##
##  StickyNote has been derived from the following work: 
##    http://mini.net/tcl/stickies
##    by Damon Courtney.
##    Dec-2004
##
##  Copyright (c) 2005 <Irrational Numbers> : <aldo.buratti@tiscali.it> 
##
##
##  This program is free software; you can redistibute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation.
##
##  Version 2.0 - ABU - November 2005


set ::THISPROG(INI)  [file join ~ .sticky.ini]
set ::THISPROG(LOCK) [file join ~ .sticky.lock]


namespace eval periodicTimer {
    namespace export schedule cancel
    
    variable ID 0
    variable SYNC_INFO ; # array
    
    proc schedule {msec args} {
        variable ID
        incr ID

        eval every $ID $msec $args
        return $ID
    }
    
    proc every {id msec args} {
        variable SYNC_INFO
        set SYNC_INFO($id) [eval \
           after $msec $args [list \; [namespace code every] $id $msec $args ]\
          ]
    }
    
    proc cancel {id} {
        variable SYNC_INFO
        after cancel $SYNC_INFO($id)
        unset SYNC_INFO($id)
    }
}



namespace eval stickyWM {
   ;
}


proc stickyWM::closeDown { iniFile } {
    if { [catch {saveIni $iniFile} ] } {
        tk_messageBox -icon error \
           -message "cannot save preferences\nin $iniFile" \
           -title "ERROR"
    }
     # remove the lock file
    catch {
     close $::lockFID
     file delete $::THISPROG(LOCK)
    }
}

proc stickyWM::resetModifiedFlag {} {
  variable somethingChanged
  foreach win [stickyNote info instances] {
    $win edit modified 0
  }   
  set somethingChanged false
}

proc stickyWM::loadIni {fname} {
    catch { source $fname }
    resetModifiedFlag
}

  # save all
  # SIDE EFFECT: reset the 'somethingChanged' flag
proc stickyWM::saveIni {fname} {
  global THISPROG

  catch {file copy -force ${fname} ${fname}.bak} 
  set f [open $fname w]
  puts $f "# ### $THISPROG(NAME) - v. $::THISPROG(VER) --- user's preferences."
  puts $f "# ### -- please do not change the following lines --"
  puts $f "wm geometry . +[winfo x .]+[winfo y .]"
  
   # save stickyNote's default
  puts $f "array set stickyNote::default [list [array get stickyNote::default]]"
   # save all stickyNotes (but stickySample) 
  global stickySample
  foreach win [stickyNote info instances] {
    if { $win == $stickySample } continue
    puts $f "set newWin \[[$win serialize]\]"
    puts $f "wm geometry \$newWin [winfo width $win]x[winfo height $win]+[winfo x $win]+[winfo y $win]"
  }
  close $f

  resetModifiedFlag   
}


   # check if someting has changed;
   #  if yes, save all 
proc stickyWM::saveIfChanged {fname} {
    variable somethingChanged

    if $somethingChanged {
        saveIni $fname
    }
}



 # create a new stickyNote close to the button "New Sticky"
 #  (having the care not to overlap the screen limits)
proc stickyWM::newSmartSticky {thisButton}  {
    set w [stickyNote create .sticky%AUTO%]
    
    set winSize(X) $stickyNote::default(width)
    set winSize(Y) $stickyNote::default(height)
        
    set screenSize(X) [winfo screenwidth $w]
    set screenSize(Y) [winfo screenheight $w]
    
    
    set p(X) [winfo pointerx $thisButton]
    set p(Y) [winfo pointery $thisButton]
    
    set b0(X) [winfo rootx $thisButton]
    set b0(Y) [winfo rooty $thisButton]
    
    set b1(X) [expr $b0(X)+[winfo width $thisButton]]
    set b1(Y) [expr $b0(Y)+[winfo height $thisButton]]
    
    # Assert: (px,py) is within the button's box 
    #         b0x<=px<=b1x
    #         b0y<=py<=b1y
    
    foreach axis {X Y} {
        set space(NEGATIVE) $b0($axis)
        set space(POSITIVE) [expr $screenSize($axis)-$b1($axis)]
        
        if { $p($axis)-$b0($axis) > $b1($axis)-$p($axis) } {
            set nearestSide POSITIVE
        } else {
            set nearestSide NEGATIVE
        }
    
        # thumb-rule
        if { $space($nearestSide) < $winSize($axis) } {
            # not enough space on this side.
            # Let's get the biggest space.
            #  NOTE: even the choosen space may be smaller than $winSize($axis)
            if { $space(NEGATIVE) > $space(POSITIVE) } {
                set nearestSide NEGATIVE
            } else {
                set nearestSide POSITIVE
            }
        }
        
        if { $nearestSide == "NEGATIVE" } {
            set B($axis) $b0($axis)
        } else {
            set B($axis) $b1($axis)
        }
        set direction($axis) $nearestSide
    }    
    
    wm geometry $w 1x1+$B(X)+$B(Y)
    update 
    $w pumpWindow 1 10  [list ${winSize(X)} ${winSize(Y)}] \
        [list $direction(X) $direction(Y)]
}


proc stickyWM::iconifyAll {}  {
    foreach tw [stickyNote info instances] {
        wm withdraw $tw
    }
}


proc stickyWM::deiconifyAll {}  {
    global stickySample
    global showSample
    
    foreach tw [stickyNote info instances] {
        if { $tw != $stickySample || $showSample } {
            wm deiconify $tw
        } 
    }
}
     



###################################################
# MAIN 
###################################################


package require Tclx

 # startup: 
 # Try to acquire a lock;
 #  if you can't, it means there's another app. running. exit
set lockFID [open $::THISPROG(LOCK) a+]
if { [flock -nowait -write $lockFID] == 0 } {
    wm withdraw .
    tk_messageBox -icon warning -title "$::THISPROG(NAME) $::THISPROG(VER)" \
        -message "$::THISPROG(NAME) is already running"
    exit
}
 # OK; you've got the lock and the rights to run .... 
 # At exit it is recommended to close it and delete it
 # Note: in case of program crash, this app can be restarted without pain,
 #       even if lock-file hasn't been deleted. 
 #       Of course, it is *better* to delete it.


package require Tk
package require stickyTpt
package require stickyNote

# WARNING: never withdraw the "." window
#  or Bwidget's SelectColor,SelectFont won't work !

wm title . "Sticky Control"
catch { wm attributes . -toolwindow 1}
wm resizable . 0 0 


 # reload saved stickies and create a sample sticky
stickyWM::loadIni  $::THISPROG(INI)
set stickySample [stickyTpt::create]



 # intercept when . is iconified/restored
bind . <Unmap> { if { "%W" == "." } { stickyWM::iconifyAll } }
bind . <Map>   { if { "%W" == "." } { stickyWM::deiconifyAll } }


 # trap <<Modified>> events (for further autosave)
bind StickyNote <<Modified>> { set stickyWM::somethingChanged true}
 # trap destroy StickyNote
bind StickyNote <Destroy> { set stickyWM::somethingChanged true }



wm protocol . WM_DELETE_WINDOW {
   stickyWM::closeDown $::THISPROG(INI)
   exit
}

 # trap logoff/shutdown.
 # Note that WM_DELETE_WINDOW handler is not called,
 #  then you should save explicitely.
 # This handler requires a non-standard support from tcltk-core
 # ( I hope it should be standard after 8.4.11 ) 
wm protocol . WM_SAVE_YOURSELF {
   stickyWM::closeDown $::THISPROG(INI)
}


 # autosave (if needed) every 5 minutes
periodicTimer::schedule [expr 5*60*1000] \
   stickyWM::saveIfChanged $::THISPROG(INI)
 


 # --- build a simple control panel ----

proc toggleSample {w val} {
   if { $val } {
       wm deiconify $w 
   } else {
       wm withdraw $w 
   }
}

. configure -bg yellow -padx 14 -pady 6

set imageDir [file join $starkit::topdir images]
set image(new) [image create photo -file [file join $imageDir "new.gif"]]
set image(settings) [image create photo -file [file join $imageDir "defaults.gif"]]

button .b1 -text "New Sticky" \
       -command "stickyWM::newSmartSticky .b1" \
       -image $image(new) -bd 3

set showSample 0
toggleSample $stickySample $showSample

checkbutton .b2 -text "Options..." -indicatoron 0 -variable showSample \
  -command { toggleSample $stickySample $showSample } \
  -selectcolor orange \
  -image $image(settings) -bd 3
  
pack .b1 .b2 -side left -expand 1 -padx 4

 # --- end of control panel


 # add support for auto-fading
package require fading

bind StickyNote <Enter> {
     transparency::fadingWindow %W  0.99 +0.1
    }
bind StickyNote <Leave> {
     transparency::fadingWindow %W [%W cget -alpha] -0.1
    }
